#!/usr/bin/perl
#
# $RCSfile: jotter.plx,v $
# $Author: swaj $
# $Revision: 1.18 $
#
# Simple PTk Jotter.
#
# Copyright (c) Alexey Semenoff 2001-2007. All rights reserved.
# Distributed under GNU Public License.
#

use 5.008;
use strict;
use warnings;
use Encode qw /encode decode from_to/;
use Getopt::Long;
use Compress::Zlib;
use Cwd;

use Tk;
use Tk::Dialog;

use vars qw /
    $toplevel
    $toplevel_sp
    $toplevel_ap
    $mw
    $menu
    $ff
    $ff2
    $ff3
    $fh
    $pw
    $fr3
    $fb
    $lab
    $lab0
    $opt
    $text
    $list
    %encode_converts
    $encode_convert
    $prev_h
    $start_from
    %config
    $PACKAGE
    $OS
    $debug
    $NAME
    $is_saved
    %DOC
    $DOC_NDX
    $file
    $update
    $crypto
    $password
    $temp_passwd
    $temp_passwd2
    $temp_passwd3
    $logo
    $img
    $VERSION
    $reask_passwd
    /;

warn 'Tested with Tk 804.027 only!' if ($Tk::VERSION < 804.027);

BEGIN {
    $PACKAGE = __PACKAGE__;

    if ( $^O =~ /^(ms)?(dos|win(32|nt)?)/i ) {
        $OS = 'WIN';
    } else {
        $OS = 'UNIX';
    }

    $_=$0;
    s|^(.+)/.*|$1|;
    $start_from = $_;

    $prev_h = 0;
    $is_saved = 1;
    $debug = 0;
    $NAME = 'Jotter';
    $is_saved = 0;
    %DOC = ();
    $DOC_NDX = 0;
    $DOC{$DOC_NDX} = q{};
    $file = q{};
    $update = 0;
    $crypto = 0;
    $password = q{};
    $toplevel_sp = q{};
    $toplevel_ap = q{};
    $temp_passwd = $temp_passwd2 = $temp_passwd3 = q{};
    $reask_passwd = 0;
    $VERSION = '1.2.4';
};

sub printd (;@);
sub printe (;@);
sub printi (;@);

use constant PROGRAM_HOME_DIR		=> '/opt/sdict';

use constant JOTTER_CONFIG_FILE		=> '.ptksdictrc';
use constant JOTTER_CONFIG_FILE_DEF	=> 'ptksdictrc.sample';
use constant JOTTER_CONFIG_FILE_WIN	=> 'ptksdictrc-win';

use constant JOTTER_FILE_SIGNATURE 	=> 'jotr';
use constant JOTTER_FILE_NOT_CRYPTED 	=> '0';
use constant JOTTER_FILE_CRYPTED 	=> '1';
use constant JOTTER_FILE_TEST_S 	=> 'asdf';

use constant JOTTER_FILE_SCRYPTED	=> 'Crypted';
use constant JOTTER_FILE_SPLAIN		=> 'Plain  ';

use constant JOTTER_FILE_SAVED 		=> '-:--';
use constant JOTTER_FILE_UNSAVED 	=> '-:**';

use constant GZIP_COMPRESSION_LEVEL 	=> 9;

binmode STDOUT, ':utf8';

#####
# S #
#####

eval 'use Crypt::Blowfish; use Digest::MD5;';

if ($@) {
    printi 'unable to use crypto:';
    printi "$@";
    $crypto = 0;
} else {
    printi 'crypto ok';
    $crypto = 1;
}


GetOptions("debug" => \$debug);

ParseConfig();

DrawInterface();

if ( ($OS eq 'UNIX') && ! $config{ 'disable_switcher' } && (exists ($config{'default_encoding'})) )
{
    if (exists ($encode_converts{ $config{ 'default_encoding' } } )) {
	$encode_convert = $config{ 'default_encoding' };
	printi "switch to '$encode_convert' encoding";
	ChangeEncoding();
    }
}

MainLoop;

####
#  #
####

sub printd (;@) {
    $debug && print "DEBUG ($PACKAGE): ", @_, "\n";
}


sub printe (;@) {
    print "ERROR ($PACKAGE): ", @_, "\n";
}


sub printi (;@) {
    print "INFO  ($PACKAGE): ", @_, "\n";
}


sub SetFileSaved {
    return unless $lab;
    $is_saved = 1;
    my $is_changed = JOTTER_FILE_SAVED;
    $lab->configure ( -text => $is_changed );
}


sub SetCrypto {
    return unless $lab0;

    if ($password eq q{} ) {
 	$lab0->configure ( -text => JOTTER_FILE_SPLAIN);
    } else { 
	$lab0->configure ( -text => JOTTER_FILE_SCRYPTED);
    }
}


sub SetFileUnSaved {
    return unless $lab;
    return unless $is_saved;
    $is_saved = 0;
    my $is_changed = JOTTER_FILE_UNSAVED;
    $lab->configure ( -text => $is_changed );
}


sub NewJotter {
    printd "NewJotter()";

    unless ($is_saved)
    {
	if ( AskFileSave() eq 'Yes' ) {
	    return unless (FileSave());
	}
    }

    CleanForNewFile();
}


sub CleanForNewFile {
    %DOC = ();
    $DOC_NDX=0;
    $DOC{$DOC_NDX} = q{};

    $file = q{};
    ClearText();
    UpdateAllList();
    UpdateTitle();

    $is_saved = 1;
    SetFileSaved();
}


sub AskFileSave {
    my $dialog = $mw->Dialog( -text => 'Quit: File changed. Save before quit?',
			      -font => $config{'font_bold'},
			      -bitmap => 'questhead',
			      -title => 'Quit',
			      -default_button => 'No',
			      -buttons => [qw/Yes No/] );
    return $dialog->Show();
}


sub CleanQuit {
    printd "CleanQuit()";

    unless ($is_saved)
    {
	if ( AskFileSave() eq 'Yes' ) {
	    return unless (FileSave());
	}
    }

    $mw->destroy if $mw;
    exit 0;
}


sub AppendEntry {
    printd "AppendEntry()";

    my $file_load = $mw->getOpenFile();

    return if ( !defined ($file_load) || ($file_load eq q{}) );

    printd "Selected file : '$file_load'";

    unless (open (IF, "< $file_load")) {
	DisplayError("Cannot open file: $!");
	return;
    }

    local $/ = undef;
    my $f = <IF>;
    close IF;

    $f =~ s|[\x00-\x08]||sg;
    $f =~ s|[\x0b-\x1f]||sg;

    if ( $config{ 'input_encoding' } )
    {
	from_to ($f, $config{ 'output_encoding' }, "utf8");
    }
    elsif ($OS eq 'UNIX' && ! $config{ 'disable_switcher' } &&
	   $encode_converts { $encode_convert } ne 'none' )
    {
	from_to ($f, $encode_converts{$encode_convert}, "utf8");
    }

    $f = decode("utf8", $f);
    $text->insert('end', $f);

    SetFileUnSaved();
    $update = 1;
}


sub AskPassword {
    printd "AskPassword()";

    return if $toplevel_ap;
    unless ($crypto) {
	DisplayError ('Crypto not found');
	return;
    }

    $toplevel_ap = $mw->Toplevel();
    $toplevel_ap->title('Set password');

    $temp_passwd3 =  $password;

    my $fr0 = $toplevel_ap->Frame ()->pack(-side =>'top',-fill => 'x');

    $fr0->Label (
		 -text => 'Enter password: ',
		 -font => $config{'font_bold'},
		 -justify => 'center' ) -> pack (
						 -side => 'left',
						 -expand => 'yes',
						 -padx => 1,
						 -pady => 1,
						 -fill => 'both'
						 );
    my $e = $fr0->Entry(
			-width => 20,
			-font => $config{'font'},
			-show => '*',
			-textvariable => \$temp_passwd3
			)-> pack(
				 -side =>'left',
				 -fill => 'x'
				 );
    $e->bind (
	      '<Return>' => [ \&OkAskPassword ]
	      );

    my $fr2 = $toplevel_ap->Frame ( ) -> pack(
					      -side => 'bottom'
					      #,-fill => 'x'
					      );
    $fr2->Button ( -text => 'Ok',
		   -borderwidth => 1,
		   -command => [ \&OkAskPassword ],
		   -padx => 10,
		   -pady => 5,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'left');

    $fr2->Button ( -text => 'Cancel',
		   -borderwidth => 1,
		   -command => [ \&CloseAskPassword ],
		   -padx => 10,
		   -pady => 5,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'right');

    CenterWindow($toplevel_ap);

    while ($toplevel_ap) {
	$mw->update if ($mw);
    }
}


sub CloseAskPassword {
    $toplevel_ap->destroy;
    $toplevel_ap = q{};
}


sub OkAskPassword {
    $password = $temp_passwd3;
    $temp_passwd3 =  q{};
	CloseAskPassword();
	SetCrypto();
}


sub ClearPassword {
    printd "ClearPassword()";
    $password = q{};
    SetCrypto();
}


sub SetPassword {
    printd "SetPassword()";

    return if $toplevel_sp;

    unless ($crypto) {
	DisplayError ('Crypto not found');
	return;
    }

    $toplevel_sp = $mw->Toplevel();
    $toplevel_sp->title('Set password');

    $temp_passwd = $temp_passwd2 = $password;

    my $fr0 = $toplevel_sp->Frame ()->pack(
					   -side =>'top',
					   -fill => 'x'
					   );

    $fr0->Label (
		 -text => 'Enter password: ',
		 -font => $config{'font_bold'},
		 -justify => 'center' ) -> pack (
						 -side => 'left',
						 -expand => 'yes',
						 -padx => 1,
						 -pady => 1,
						 -fill => 'both'
						 );
    $fr0->Entry(
		-width => 20,
		-font => $config{'font'},
		-show => '*',
		-textvariable => \$temp_passwd
		)-> pack(
			 -side =>'right',
			 -fill => 'x',
			 -padx => 1,
			 -pady => 1
			 );

    my $fr1 = $toplevel_sp->Frame ()->pack(-side =>'top',-fill => 'x');

    $fr1->Label (
		 -text => 'Enter password (again): ',
		 -font => $config{'font_bold'},
		 -justify => 'center' ) -> pack (
						 -side => 'left',
						 -expand => 'yes',
						 -padx => 1,
						 -pady => 1,
						 -fill => 'both'
						 );
    $fr1->Entry(
		-width => 20,
		-font => $config{'font'},
		-show => '*',
		-textvariable => \$temp_passwd2
		)-> pack(
			 -side =>'right',
			 -fill => 'x',
			 -padx => 1,
			 -pady => 1
			 );

    my $fr2 = $toplevel_sp->Frame ( ) -> pack(
					      -side => 'bottom'
					      #,-fill => 'x'
					      );

    $fr2->Button ( -text => 'Ok',
			 -borderwidth => 1,
			 -command => [ \&OkSetPassword ],
			 -padx => 10,
			 -pady => 5,
			 -font => $config{'font_bold'},
			 -underline => 0,
			 ) -> pack (-side => 'left');



    $fr2->Button ( -text => 'Cancel',
		   -borderwidth => 1,
		   -command => [ \&CloseSetPassword ],
		   -padx => 10,
		   -pady => 5,
		   -font => $config{'font_bold'},
		   -underline => 0
		   ) -> pack (-side => 'right');

    CenterWindow($toplevel_sp);
}


sub CloseSetPassword {
    $toplevel_sp->destroy;
    $toplevel_sp = q{};
}


sub OkSetPassword {

    if ($temp_passwd ne $temp_passwd2) {
	DisplayError("Passwords mismatch!");
    } else {
	$password = $temp_passwd;
	$temp_passwd = $temp_passwd2 = q{};
	CloseSetPassword();
	SetCrypto();
    }
}


sub PrintHelp {
    printd "PrintHelp()";

    $fh->configure ( -state => 'disabled' );
    $toplevel = $mw->Toplevel();
    $toplevel->title('Help');

    my $fr = $toplevel->Frame () -> pack (
					  -side =>'top',
					  -fill => 'x',
					  );

    my $text = q{};

    unless ($crypto) {
	$text = <<EOS;

PTk Jotter. Version: $VERSION ($OS).

Written by Alexey Semenoff [http://swaj.net] (c) 2001-2007.

Perl: $]; Compress::Zlib: $Compress::Zlib::VERSION, Tk: $Tk::VERSION.

EOS
	} else {
	$text = <<EOS;

PTk Jotter. Version: $VERSION ($OS).

Written by Alexey Semenoff [http://swaj.net] (c) 2001-2007.

Perl: $]; Compress::Zlib: $Compress::Zlib::VERSION, Tk: $Tk::VERSION, MD5: $Digest::MD5::VERSION, Blowfish: $Crypt::Blowfish::VERSION.

EOS
	}

    $fr ->Label (
		 -image => $img
		 ) -> pack(
			   -side => 'left',
			   -padx => 15,
			   -pady => 15
			   );

    $fr->Label ( -text => $text,
		 -font => $config{'font_bold'},
		 -justify => 'center' ) -> pack (
						 -side => 'top',
						 -expand => 'yes',
						 -padx => 10,
						 -pady => 10,
						 -fill => 'both'
						 );

    my $fr2 = $toplevel->Frame () -> pack (
					   -side => 'bottom',
					   -fill => 'x',
					   );

    $fr2->Button ( -text => 'Continue',
		   -borderwidth => 1,
		   -command => [ \&ClosePrintHelp ],
		   -padx => 20,
		   -pady => 10,
		   -font => $config{'font_bold'},
		   -underline => 0 ) -> pack ( -side => 'bottom');

    CenterWindow($toplevel);
}


sub ClosePrintHelp {
    $toplevel->destroy;
    $fh->configure ( -state => 'normal' );
}


sub CenterWindow {
    my $w = shift;
    return unless $w;
    $mw->update;
    $_ = $pw->geometry;
    printd "pw geometry $_";
    @_ = split /[x\+]/;
    my ($m_w, $m_h) = @_[0,1];
    $_ = $mw->geometry;
    printd "mw geometry $_";
    @_ = split /[x\+]/;
    my ($m_x, $m_y) = @_[2,3];
    $_ = $w->geometry;
    @_ = split /[x\+]/;
    my ($t_x, $t_y) = @_[0,1];
    my $geom = $t_x . 'x' . $t_y . '+' . int ( ($m_w  - $t_x) /2 + $m_x ) . '+' . int ( ($m_h - $t_y) / 2 + $m_y );
    $w->geometry($geom);
}

sub SaveCurrentTxt {
    printd "SaveCurrentTxt ()";

    my $file_export = $mw->getSaveFile();

    return if (!defined($file_export)||($file_export eq q{}));

    printd "Selected file : '$file_export'\n";

    unless (open (OF, "> $file_export")) {
	DisplayError("Cannot export file: $!");
	return ;
    }

    my $t = $text->get('1.0', 'end');

    $t = encode("utf8", $t);

    if ( $config{ 'output_encoding' } )
    {
	from_to ( $t, "utf8", $config{ 'output_encoding' } );
    }
    elsif ( ! $config{ 'disable_switcher' } && $encode_converts { $encode_convert } ne 'none' )
    {
	from_to ( $t, "utf8",  $encode_converts { $encode_convert } );
    }

    print OF $t;
    close OF;

    printd "done";
}

sub ExportTxt {
    printd "ExportTxt()";

    my $file_export = $mw->getSaveFile();

    return if (!defined($file_export)||($file_export eq q{}));

    printd "Selected file : '$file_export'\n";

    UpdateCurList();

    unless (open (OF, "> $file_export")) {
	DisplayError("Cannot export file: $!");
	return ;
    }

    print OF <<EOS;
#
# Created by $0
#
<header>
title = <Edit> dictionary
copyright = Converted by <Edit> under GNU Public License
version = 1.0
w_lang = <Edit>
a_lang = <Edit>
</header>
#
# Begin of articles
#
EOS

    my ($k, $v, $doc);

    for (sort {$a <=> $b } keys (%DOC))
    {
	$doc = $DOC{$_};
	$doc = encode("utf8", $doc);

	@_ = split (/\n/, $doc, 2);

	($k, $v) = @_;

	$k = 'Empty' if ($k eq q{});

	$v =~ s|\n\n|<p>|isg;
	$v =~ s|[\x00-\x1f]| |sg;
	$v =~ s|\s\s| |isg;
	$v = 'Empty' if ($v eq q{});

	print OF $k , '___' , $v , "\n";
    }

    print OF <<EOS;
#
# End of articles
#
EOS

    close OF;
}


sub ImportTxt {
    printd "ImportTxt()";

    unless ($is_saved) {
	DisplayError('Current file unsaved. Save first!');
	return;
    }

    #my @types = ( ["All files"  , '*' ] );
    my $file_load = $mw->getOpenFile(); # -filetypes => \@types );

    return if (!defined($file_load)||($file_load eq q{}));

    printd "Selected file : '$file_load'\n";

    unless (open (IF, "< $file_load")) {
	DisplayError("Cannot open file: $!");
	return;
    }

    %DOC = ();
    $DOC_NDX=0;

    my ($k, $v, $unit);

    while (<IF>) {
	s|\r||s;
	chomp;
	next if /^\#/;

	($k, $v) = split (/___/, $_, 2);

	unless ($k && $v) {
	    printi "skipped line '$_'\n";
	    next;
	}

	$unit = $k . "\n" . $v;

	$unit = decode("utf8", $unit);

	printd "unit = $unit";

	$DOC{$DOC_NDX++} = $unit;
    }

    close IF;

    $DOC_NDX=0;
    SelectEntry();
    UpdateAllList();
    UpdateTitle();
}


sub FileOpen {
    printd "FileOpen()";

    unless ($is_saved) {
	DisplayError('Current file unsaved. Save first!');
	return;
    }

    my $cwd = cwd();

    if (exists ($config{'jotter_dir'})) {
	chdir ($config{'jotter_dir'});
    }

    #my @types = ( ["All files"  , '*' ] );
    my $file_load = $mw->getOpenFile(); # -filetypes => \@types );

    chdir ($cwd);

    return if (!defined($file_load)||($file_load eq q{}));

    printd "Selected file : '$file_load'\n";

    unless (open (IF, "< $file_load")) {
	DisplayError("Cannot open file: $!");
	return;
    }
    binmode (IF);

    my $size = (stat(IF))[7];

    my $hdr = q{};
    my $crypted = q{};

    unless (sysread (IF, $hdr, 4, 0)) {
        LoadJotterError();
	close IF;
        return;
    }

    printd "sign = $hdr";

    unless ($hdr eq JOTTER_FILE_SIGNATURE) {
	printd "wrong signature";
        LoadJotterError();
	close IF;
        return;
    }

    unless (sysread (IF, $crypted, 1, 0)) {
        LoadJotterError();
	close IF;
        return;
    }

    printd "crypted = $crypted";

    my $cdata = q{};

    unless (sysread (IF, $cdata, $size - 5, 0)) {
        LoadJotterError();
	close IF;
        return;
    }
    close IF;


    if ($crypted eq '1')
    {
	if (($password eq q{}) || $reask_passwd) {
	    AskPassword();
	    if ($password eq q{}) {
		LoadJotterError();
		return;
	    }
	}

	$reask_passwd = 1;

	my $digest = Digest::MD5::md5($password);

	my $cipher = new Crypt::Blowfish ($digest);

	my $rlen = length($cdata);
	my $pad = $rlen % 8;
	my $blocks = $rlen / 8;

	printi "crypted len: $rlen; pad: $pad; blocks: $blocks " ;

	if ($pad) {
	    DisplayError ('wrong crypted size');
	    return;
	}


	my $blk;
	my $en_cdata = q{};

	for (my $i=0;  $i < $blocks; $i++)  {

	     $blk = $cipher->decrypt(substr($cdata, $i*8, 8));

	     if (length ($blk) == 0) {
		 DisplayError("Cannot encrypt block $i (blowfish error)");
		 return ;
	     }

	     $en_cdata .= $blk;
   	 }

	 $cdata = $en_cdata;
    }
    else { # not crypted
	$password = q{}; # clear the password
    }

    my $data = uncompress ($cdata);

    SetCrypto();

    unless ($data) {
        LoadJotterError();
	close IF;

	if ($crypted eq '1')
	{
	    $password = q{}; # clear the password

	}
	SetCrypto();
        return;
    }

    my $tst = substr ($data, 0, 4);

    printd "tst = $tst";

    if ($tst ne JOTTER_FILE_TEST_S) {
        LoadJotterError();
	close IF;
        return;
    }

    my $offset = 4;
    my $len = 0;
    my $unit = q{};

    %DOC = ();
    $DOC_NDX=0;

    while (1) {
	$len = unpack ("L", substr($data, $offset, 4));
	last if ($len == 0);
	$offset += 4;
	printd "len = $len";

	$unit = substr ($data, $offset, $len);
	$offset += $len;

	$unit = decode("utf8", $unit);

	printd "unit = $unit";

	$DOC{$DOC_NDX++} = $unit;
    }

    $file =  $file_load;

    $DOC_NDX=0;
    SelectEntry();

    UpdateAllList();

    UpdateTitle();

    $reask_passwd = 0;
}


sub FileClose {
    printd "FileClose()";
    NewJotter();
}


sub FileSaveAs {
    printd "FileSaveAs()";

    my $cwd = cwd();

    if (exists ($config{'jotter_dir'})) {
	chdir ($config{'jotter_dir'});
    }

    #my @types = ( ["All files"  , '*' ] );
    my $file_save = $mw->getSaveFile(); # -filetypes => \@types );

    chdir ($cwd);

    return if (!defined($file_save)||($file_save eq q{}));

#    $file_save .= '.jot' if ($file_save !~ /\.jot/);

    printd "Selected file : '$file_save'\n";
    $file = $file_save;

    SaveJotter();
}


sub FileSave {
    printd "FileSave()";

    if ($file eq q{}) {
	FileSaveAs();
    } else {
	SaveJotter();
    }
}


sub DisplayError {
    my $text = shift;

    $text = 'Error: ' . $text;

    my $dialog = $mw->Dialog( -text => $text,
			      -font => $config{'font_bold'},
			      -bitmap => 'error',
			      -title => 'Error',
			      -default_button => 'Continue',
			      -buttons => [qw/Continue/] );
    my $d = $dialog->Show();
}


sub LoadJotterError {
    DisplayError ('Cannot load file');
}


sub SaveJotterError {
    DisplayError('Internal, cannot save file');
}


sub SaveJotter {
    printd "SaveJotter()  '$file'";

    UpdateCurList();

    unless (open (OF, "> $file")) {
	DisplayError("Cannot save file: $!");
	return 0;
    }

    binmode (OF);

    syswrite (OF, JOTTER_FILE_SIGNATURE);

    if ($crypto && $password ne q{}) {
	syswrite (OF, JOTTER_FILE_CRYPTED);
    } else {
	syswrite (OF, JOTTER_FILE_NOT_CRYPTED);
    }

    my $data = JOTTER_FILE_TEST_S;
    my $unit;
    my $doc;
    for (sort {$a <=> $b } keys (%DOC)) {
	$unit = undef;
	$doc = $DOC{$_};
	$doc = encode("utf8", $doc);
	$unit = pack ("L", length($doc));
	printd ">$doc<";
	$unit .= $doc;
	$data .= $unit;
    }

    $data .= pack ("L", 0);

    my $cdata = compress ($data, GZIP_COMPRESSION_LEVEL);

    if ($crypto && $password ne q{}) {

	 my $digest = Digest::MD5::md5($password);

	 if (length ($digest) != 16) {
	     DisplayError("Cannot save file (md5 length is wrong");
	     close OF;
	     return 0;
	 }

	 my $dlen = length ($cdata);
	 printi "data len = $dlen";

	 my $blocks = int ($dlen/8); 
	 my $pad = $dlen % 8;

	 if ($pad != 0) {
	     $cdata .= "\x0" x (8 - $pad);
	     $blocks++;
	     $dlen = length ($cdata);
	     printi "length padded";
	 }

	 printi "actual length = $dlen, ($blocks block(s))";

	 my $en_cdata = q{};

	 my $cipher = new Crypt::Blowfish ($digest);

	 my $blk;

	 for (my $i=0;  $i < $blocks; $i++)  {

	     $blk = $cipher->encrypt(substr($cdata, $i*8, 8));

	     if (length ($blk) == 0) {
		 DisplayError("Cannot encrypt block $i (blowfish error)");
		 close OF;
		 return 0;
	     }

	     $en_cdata .= $blk;
   	 }

	 $cdata = $en_cdata;
     }

    syswrite (OF, $cdata);
    close OF;

    SetFileSaved();
    UpdateTitle();
    return 1;
}


sub ChangeEncoding {
    printd "ChangeEncoding()";
    printd "Use encoding (main) '$encode_convert'";
}


sub NewEntry {
    printd "NewEntry()";

    UpdateCurList();

    my @ks = sort { $a <=> $b } keys (%DOC);

    my $e = $#ks;

    printd "max element = $e";

    $DOC_NDX = $ks[$e] +1;

    $DOC{$DOC_NDX} = q{};

    ClearText();

    UpdateAllList();

    printd "allocated new entry '$DOC_NDX'";
}


sub GetFirstEntry {
    for (sort { $a <=> $b } keys (%DOC)) {
	printd "first entry: $_";
	return $_;
    }
    return 0;
}


sub ClearText {
    return unless $text;
    $text->delete('1.0', 'end');
}


sub DeleteEntry {
    printd "DeleteEntry()";
    my $sz = scalar (keys(%DOC));

    printd "items: $sz, IDX = $DOC_NDX";

    ClearText();

    if ($sz > 1) {
	delete $DOC{$DOC_NDX};
	$DOC_NDX = GetFirstEntry();
	$text->insert('1.0', $DOC{$DOC_NDX});
    } else {
	$DOC{$DOC_NDX} = q{};

    }

    UpdateAllList();
}


sub UpdateAllList {
    printd "UpdateAllList()";
    my @list = ();

    $list->delete(0, 'end');

    for (sort { $a <=> $b } keys (%DOC)) {
	my $it = $DOC{$_};
	$it =~ s|^(.+?)\n.*|$1|is;

	$it = '* Empty *' if ($it eq q{});
	push @list, $it;
	# printd "added $it to listbox";
    }

    $list->insert(0, @list);
    $list->activate($DOC_NDX);
}


sub GetCurrentListBoxId {
    my $i = 0;

    for (sort {$a <=> $b } keys (%DOC)) {
	return $i if ($_ == $DOC_NDX);
	$i++;
    }
    return $i;
}


sub UpdateCurList {
    printd "UpdateCurList()";

    my $i = GetCurrentListBoxId();

    printd "updating entry in listbox: $i";

    $list->delete($i);

    my $t = $text->get('1.0', 'end');
    $DOC{$DOC_NDX} = $t;

    my $it = $t;
    $it =~ s|^(.+?)\n.*|$1|is;
    $it = '* Empty *' if ($it eq q{});

    $list->insert($i, $it);
}


sub SelectEntry {
    printd "SelectEntry()";
    ClearText();
    $text->insert('1.0', $DOC{$DOC_NDX});
}


sub ListSel {
    printd "ListSel()";

    my $c = shift;

    printd "selected listbox item: '$c'";

    my @ks = sort { $a <=> $b } keys (%DOC);

    $DOC_NDX = $ks[$c];
    SelectEntry();
}


sub UpdateTitle {
    $mw->title("$NAME: $file");
}


sub UpdateTimer {
    # printd 'UpdateTimer()';
    return unless $update;
    UpdateCurList();
    $update = 0;
}


sub DrawInterface {
    $mw = MainWindow->new;
    $mw->title($NAME);
    $mw->CmdLine;
    $mw->protocol('WM_DELETE_WINDOW' => [\&CleanQuit ]);

    $mw->repeat (1000 => sub { UpdateTimer(); });

    $menu = $mw->Frame (
			-relief => 'raised',
			-borderwidth => 1,
			) -> pack (
				   -fill =>'x',
				   );

    $logo .= $_ while (<DATA>);

    $img = $menu->Photo(
			-format => 'xpm',
			-data  => $logo
			);

    $debug || $mw->Icon(
			-image => $img
			);

    $ff = $menu->Menubutton (
			     -text => 'File', 
                             -font => $config{'font_bold'},
                             -tearoff => 0,
                             -padx => 1,
                             -pady => 1,
                             -underline => 0
			     )-> pack(
				      -side => 'left'
				      );
    $ff->command (
		  -label => 'New',
		  -font => $config{'font_bold'},
		  -command => [ \&NewJotter],
		  -underline => 0
		  );

    $ff->command (
		  -label => 'Open...',
		  -font => $config{'font_bold'},
		  -command => [ \&FileOpen],
		  -underline => 0
		  );

    $ff->command (
		  -label => 'Save',
		  -font => $config{'font_bold'},
		  -command => [ \&FileSave],
		  -underline => 0
		  );

    $ff->command (
		  -label => 'Save As...',
		  -font => $config{'font_bold'},
		  -command => [ \&FileSaveAs],
		  -underline => 0
		  );

    $ff->command (
		  -label => 'Close',
		  -font => $config{'font_bold'},
		  -command => [ \&FileClose],
		  -underline => 0
		  );



    $ff->command (
		  -label => 'Quit',
		  -font => $config{'font_bold'},
		  -command => [ \&CleanQuit],
		  -underline => 0
		  );

    $ff2 = $menu->Menubutton (
			     -text => 'Edit',
                             -font => $config{'font_bold'},
                             -tearoff => 0,
                             -padx => 10,
                             -pady => 1,
                             -underline => 0
			     )-> pack(
				      -side => 'left'
				      );

    $ff2->command (
		  -label => 'New Entry',
		  -font => $config{'font_bold'},
		  -command => [ \&NewEntry],
		  -underline => 0
		  );

    $ff2->command (
		  -label => 'Delete Entry',
		  -font => $config{'font_bold'},
		  -command => [ \&DeleteEntry],
		  -underline => 0
		  );

    $ff2->command (
		  -label => 'Append from text file...',
		  -font => $config{'font_bold'},
		  -command => [ \&AppendEntry],
		  -underline => 0
		  );

    my $bp = $ff2->command (
			    -label => 'Set password...',
			    -font => $config{'font_bold'},
			    -command => [ \&SetPassword],
			    -underline => 0
			    );

    my $bp2 = $ff2->command (
			     -label => 'Clear password',
			     -font => $config{'font_bold'},
			     -command => [ \&ClearPassword],
			     -underline => 0
			     );

    $bp->configure ( -state => 'disabled' ) unless ($crypto);
    $bp2->configure ( -state => 'disabled' ) unless ($crypto);

    $fh = $menu->Button (
			  -text => 'Help',
			  -font => $config{'font_bold'},
			  -borderwidth => 0,
			  -padx => 1,
			  -pady => 1,
			  -command => [ \&PrintHelp ],
			  -underline => 0
			  ) -> pack( -side => 'right' );


    $ff3 = $menu->Menubutton (
			     -text => 'Tools', 
                             -font => $config{'font_bold'},
                             -tearoff => 0,
                             -padx => 10,
                             -pady => 1,
                             -underline => 0
			     )-> pack(
				      -side => 'left'
				      );

#    $ff3->command (
#		  -label => 'Search...',
#		  -font => $config{'font_bold'},
#		  -command => [ \&Search],
#		  -underline => 0
#		  );

    $ff3->command (
		  -label => 'Export to text/sdct...',
		  -font => $config{'font_bold'},
		  -command => [ \&ExportTxt],
		  -underline => 0
		  );


    $ff3->command (
		  -label => 'Import from text/sdct...',
		  -font => $config{'font_bold'},
		  -command => [ \&ImportTxt],
		  -underline => 0
		  );

    $ff3->command (
		  -label => 'Save current note to file...',
		  -font => $config{'font_bold'},
		  -command => [ \&SaveCurrentTxt],
		  -underline => 0
		  );

    $pw = $mw->Panedwindow;
    $pw->pack(qw/-side top -expand yes -fill both -pady 0 -padx 0/);


    $text = $pw->Scrolled(
                           qw/Text -scrollbars e -wrap word/
                           )->pack(
                                   -fill => 'both',
                                   -expand => 'yes'
                                   );

    $text->configure (
                      -font => $config{'font'}
                      );

    $text->bind (
		 '<Key>' => [ \&TextPress ]
		 );

    if ( $OS eq 'UNIX' ) {
      $text->bind (
		   '<ButtonRelease-2>' => [ \&TextPaste ]
		  );
    }
    else {
      $text->bind (
		    '<Control-v>' => [ \&TextPaste ]
		   );
    }

    $list = $pw->Scrolled(
			   qw/Listbox -scrollbars e -width 20 -height 10 -setgrid 1/

			  )->pack(
				   -fill => 'both',
				   -expand => 'yes'
				   );

    $list->configure (
		      -font => $config{'font'}
		      );

    $list->bind(
		'<<ListboxSelect>>' => sub { ListSel ( ($list->curselection)[0] ) }
		);

    $list->activate(0);
    $list->focus();

    UpdateAllList();

    $pw->add($text, $list);

    $fb = $mw->Frame (
		      -relief => 'raised',
		      -borderwidth => 1
		      ) -> pack (
				 -side => 'bottom',
				 -fill =>'x'
				 );


    $lab0 = $fb->Label (
			-font => $config{'font'},
			-justify => 'center' ) -> pack (
							-side => 'left',
							-expand => 'no',
							-padx => 1,
							-pady => 1,
							-fill => 'both'
							);
    $lab = $fb->Label (
		       -font => $config{'font'},
			-justify => 'center' ) -> pack (
							-side => 'left',
							-expand => 'no',
							-padx => 1,
							-pady => 1,
							-fill => 'both'
							);
    SetFileSaved();
    SetCrypto();

    if ($OS eq 'UNIX' && ! $config{ 'disable_switcher' } )
    {
	$opt = $fb->Optionmenu(
			       -options => [ () ] ,
			       -variable => \$encode_convert,
			       -command => [ \&ChangeEncoding ]
			       ) -> pack(
					 -side => 'left'
					 );
	$opt->addOptions (
			  sort keys (%encode_converts)
			  );
    }

    $mw->update if ($mw);
}


sub TextPaste {
    printd "TextPaste()";
    SetFileUnSaved();
    $update = 1;
}


sub TextPress {
    printd "TextPress()";

    SetFileUnSaved();

    if ($OS eq 'WIN' || $config{ 'disable_switcher' } )
    {
	$update = 1;
	return;
    }

    my($widget) = @_;
    my $e = $widget->XEvent;    # get event object
    my($keysym_text, $keysym_decimal) = ($e->K, $e->N);
    printd "pressed: '$keysym_text' dec: '$keysym_decimal' ";

    # if ($keysym_text eq 'Return') { printd "RET:"; $text->openLine; return; }

    my $last = $text->get('end - 2 chars' , 'end' );
    chomp ($last); 
    printd "last = '$last'";

    my $h = ord ($last);

    printd "ord = '$h'";

    if ($h < 128 || $h > 255) {
	$prev_h = $h;
	$update = 1;
	return;
    }

    my $res = $text->search(-backwards => "$last", 'end');
    printd "res = '$res'";
    $text->delete ( $res , 'end' );

    from_to ($last, $encode_converts{$encode_convert}, "utf8");

    $last = decode("utf8", $last);

    if ($prev_h == 10) {
	$text->insert ('end', "\n");
    }

    $text->insert ('end', $last);
    $prev_h = 0;
    $update = 1;
}


sub ParseConfig {
    my $file;
    my %tempcfg = ();
    my @temp = ();
    my @temp2 = ();

    if ($OS eq 'WIN') {
	$file = $start_from . '/../etc/' . JOTTER_CONFIG_FILE_WIN;
    } else { 
	$file = "$ENV{'HOME'}/" . JOTTER_CONFIG_FILE;

	unless (-e $file) {
	    my $o = $file;
	    $file = PROGRAM_HOME_DIR . '/etc/' . JOTTER_CONFIG_FILE_DEF;
	    printi "rc file '$o' not found, using global '$file' instead";
	}
    }

    if (open (F, "< $file")) {
	while (<F>) {
	    chomp;
	    s/\r$//g;
	    next if /^\#/;
	    next if /^\s*$/;
	    my ($p,$v)=(split(/\s=\s/, $_, 2))[0,1];
	    $p =~ s|^\s+||g;
	    $p =~ s|\s+$||g;
	    $v =~ s|^\s+||g;
	    $v =~ s|\s+$||g;
	    next unless ($p && $v);
	    $tempcfg{$p} = $v;
	}
	close F;
    } else {
	printd "Unable read configuration from file '$file'";
    }

    @temp = ();
    @temp = split /\s+/, $tempcfg{'input_encodings'} if ( $tempcfg{'input_encodings'} ); 

    if ( @temp )
    {
	$config{ 'input_encodings' } = \@temp;
	$config{ 'disable_switcher' } = 0 ;
    } else {
	$config{ 'input_encodings' } = [ ] ; #  [ qw / rus koi8-r lat iso8859-1/ ];
	$config{ 'disable_switcher' } = 1 ;
    }

    $config{'font'} = $tempcfg{'font'} || '-*-*-*-*-*--*-*-*-*-*-*-iso10646-*';

    unless (exists ($tempcfg{'font_bold'})) {
	$config{'font_bold'} = $config{'font'};
	$config{'font_bold'} =~ s|^(\-.+?\-.+?)\-.+?\-(.*)$|$1-bold-$2|;
    } else {
	$config{'font_bold'} = $tempcfg{'font_bold'};
    }

    if (exists ($tempcfg{'jotter_dir'})) {
	$config{'jotter_dir'} = $tempcfg{'jotter_dir'};
    }

    if (exists ($tempcfg{'default_encoding'})) {
	$config{'default_encoding'} = $tempcfg{'default_encoding'};
    }

    if (exists ($tempcfg{ 'output_encoding' }) ) {
	$config{ 'output_encoding' } = $tempcfg{ 'output_encoding' };
    }

    if (exists ($tempcfg{ 'input_encoding' }) ) {
	$config{ 'input_encoding' } = $tempcfg{ 'input_encoding' };
    }

    %encode_converts = @{ $config{ 'input_encodings' } };

    # print Dumper %config ; exit;
    return 1;
}

__DATA__
/* XPM */
static char *logo_jotter_32x32_16[] = {
/* columns rows colors chars-per-pixel */
"32 32 16 1",
"  c #1A1715",
". c #0F3A58",
"X c #214D6B",
"o c #503019",
"O c #664B31",
"+ c #626667",
"@ c #366383",
"# c #57758A",
"$ c #6E8799",
"% c #8D6E4B",
"& c #AB8B64",
"* c #99A0A5",
"= c #9CB3C3",
"- c #C9BFB2",
"; c #D5D5D5",
": c #FEFEFE",
/* pixels */
"::::::=============$*====;::::::",
"::::::==*#XXXXXX.XXX.X#$*:::::::",
":::::::::::$XXXX...X-:::::::::::",
"::::::::::::@@XX..X$::::::::::::",
"::::::::::::$@XX...-::::::::::::",
"::::::::::::=@@X...;::::::::::::",
"::::::::::::=@@XX.X;::::::::::::",
"::::::::::::=@@@XXX;::::::::::::",
"::::::::::::=@@X+XX;::::::::::::",
"::::::::::::=##@XXX;::::::::::::",
"::::::::::::=$#@+XX;::::::::::::",
"::::::::::::=###XO+;::::::::::::",
"::::::::::::=$$#+X+;::::::::::::",
"::::::::::::;$$##++;::::::::::::",
"::::::::::::=*$$#+#;::::::::::::",
"::::::::::::;**$#+#;::::::::::::",
"::::::::::::;=**#+$;::::::::::::",
"::::::::::::;--**+&;::::::::::::",
"::::::::::::**&&*++;::::::::::::",
"::::::::::::-OOOooo;::::::::::::",
"::::::::::::-&&&%O%;::::::::::::",
"::::::::::::-&&&%O%;::::::::::::",
"::::::::::::-&&&O+O;::::::::::::",
"::::::::::::-&%&O%O:::::::::::::",
"::::::::::::-%%%oO+:::::::::::::",
"::::::::::::&%%OoO&:::::::::::::",
"::;--;::::::-O%oOO=;;;;;;;:;;:::",
":&Oooo&:::::&OoooOo     +*;;;;;:",
";oooooo*::::+Ooo.      +;:::::::",
";oo   oo-::;Ooo      o*;::::::::",
":*ooooooo%*Oo o     +;::::::::::",
":;&OoooooooO+&+**-;:::::::::::::"
};
__END__
